@@ -1,3 +1,5 @@
+"""The streamrip command line interface."""
+
 import logging
 import os
 from getpass import getpass
@@ -6,13 +8,15 @@ from hashlib import md5
 import click
 import requests
 
+from . import __version__
 from .clients import TidalClient
 from .config import Config
 from .constants import CACHE_DIR, CONFIG_DIR, CONFIG_PATH, QOBUZ_FEATURED_KEYS
 from .core import MusicDL
-from .utils import init_log
 
-logger = logging.getLogger(__name__)
+
+logging.basicConfig(level="WARNING")
+logger = logging.getLogger("streamrip")
 
 if not os.path.isdir(CONFIG_DIR):
     os.makedirs(CONFIG_DIR, exist_ok=True)
@@ -21,9 +25,14 @@ if not os.path.isdir(CACHE_DIR):
 
 
 @click.group(invoke_without_command=True)
-@click.option("-c", "--convert", metavar="CODEC", help="alac, mp3, flac, or ogg")
 @click.option(
-    "-u", "--urls", metavar="URLS", help="Url from Qobuz, Tidal, SoundCloud, or Deezer"
+    "-c", "--convert", metavar="CODEC", help="alac, mp3, flac, or ogg"
+)
+@click.option(
+    "-u",
+    "--urls",
+    metavar="URLS",
+    help="Url from Qobuz, Tidal, SoundCloud, or Deezer",
 )
 @click.option(
     "-q",
@@ -34,6 +43,7 @@ if not os.path.isdir(CACHE_DIR):
 @click.option("-t", "--text", metavar="PATH")
 @click.option("-nd", "--no-db", is_flag=True)
 @click.option("--debug", is_flag=True)
+@click.version_option(prog_name="streamrip")
 @click.pass_context
 def cli(ctx, **kwargs):
     """Streamrip: The all-in-one Qobuz, Tidal, SoundCloud, and Deezer music downloader.
@@ -51,7 +61,8 @@ def cli(ctx, **kwargs):
     global core
 
     if kwargs["debug"]:
-        init_log()
+        logger.setLevel("DEBUG")
+        logger.debug("Starting debug log")
 
     config = Config()
 
@@ -59,13 +70,9 @@ def cli(ctx, **kwargs):
         return
 
     if config.session["check_for_updates"]:
-        from importlib import metadata
-
-        from packaging import version
-
         r = requests.get("https://pypi.org/pypi/streamrip/json").json()
         newest = r["info"]["version"]
-        if version.parse(metadata.version("streamrip")) < version.parse(newest):
+        if __version__ != newest:
             click.secho(
                 "A new version of streamrip is available! "
                 "Run `pip3 install streamrip --upgrade` to update.",
@@ -124,7 +131,6 @@ def filter_discography(ctx, **kwargs):
 
     For basic filtering, use the `--repeats` and `--features` filters.
     """
-
     filters = kwargs.copy()
     filters.pop("urls")
     config.session["filters"] = filters
@@ -134,9 +140,14 @@ def filter_discography(ctx, **kwargs):
 
 
 @cli.command()
-@click.option("-t", "--type", default="album", help="album, playlist, track, or artist")
 @click.option(
-    "-s", "--source", default="qobuz", help="qobuz, tidal, soundcloud, or deezer"
+    "-t", "--type", default="album", help="album, playlist, track, or artist"
+)
+@click.option(
+    "-s",
+    "--source",
+    default="qobuz",
+    help="qobuz, tidal, soundcloud, or deezer",
 )
 @click.argument("QUERY", nargs=-1)
 @click.pass_context
@@ -178,7 +189,7 @@ def search(ctx, **kwargs):
 @click.option("-l", "--list", default="ideal-discography")
 @click.pass_context
 def discover(ctx, **kwargs):
-    """Searches for albums in Qobuz's featured lists.
+    """Search for albums in Qobuz's featured lists.
 
     Avaiable options for `--list`:
 
@@ -224,12 +235,14 @@ def discover(ctx, **kwargs):
 
 @cli.command()
 @click.option(
-    "-s", "--source", help="Qobuz, Tidal, Deezer, or SoundCloud. Default: Qobuz."
+    "-s",
+    "--source",
+    help="Qobuz, Tidal, Deezer, or SoundCloud. Default: Qobuz.",
 )
 @click.argument("URL")
 @click.pass_context
 def lastfm(ctx, source, url):
-    """Searches for tracks from a last.fm playlist on a given source.
+    """Search for tracks from a last.fm playlist on a given source.
 
     Examples:
 
@@ -241,7 +254,6 @@ def lastfm(ctx, source, url):
 
         Download a playlist using Tidal as the source
     """
-
     if source is not None:
         config.session["lastfm"]["source"] = source
 
@@ -251,16 +263,30 @@ def lastfm(ctx, source, url):
 
 @cli.command()
 @click.option("-o", "--open", is_flag=True, help="Open the config file")
-@click.option("-d", "--directory", is_flag=True, help="Open the config directory")
+@click.option(
+    "-d", "--directory", is_flag=True, help="Open the config directory"
+)
 @click.option("-q", "--qobuz", is_flag=True, help="Set Qobuz credentials")
 @click.option("-t", "--tidal", is_flag=True, help="Re-login into Tidal")
 @click.option("--reset", is_flag=True, help="RESET the config file")
+@click.option(
+    "--update",
+    is_flag=True,
+    help="Reset the config file, keeping the credentials",
+)
+@click.option("-p", "--path", is_flag=True, help="Show the config file's path")
 @click.pass_context
 def config(ctx, **kwargs):
     """Manage the streamrip configuration file."""
     if kwargs["reset"]:
         config.reset()
 
+    if kwargs["update"]:
+        config.update()
+
+    if kwargs["path"]:
+        print(CONFIG_PATH)
+
     if kwargs["open"]:
         click.secho(f"Opening {CONFIG_PATH}", fg="green")
         click.launch(CONFIG_PATH)
@@ -271,7 +297,9 @@ def config(ctx, **kwargs):
         click.launch(config_dir)
 
     if kwargs["qobuz"]:
-        config.file["qobuz"]["email"] = input(click.style("Qobuz email: ", fg="blue"))
+        config.file["qobuz"]["email"] = input(
+            click.style("Qobuz email: ", fg="blue")
+        )
 
         click.secho("Qobuz password (will not show on screen):", fg="blue")
         config.file["qobuz"]["password"] = md5(
@@ -279,7 +307,9 @@ def config(ctx, **kwargs):
         ).hexdigest()
 
         config.save()
-        click.secho("Qobuz credentials hashed and saved to config.", fg="green")
+        click.secho(
+            "Qobuz credentials hashed and saved to config.", fg="green"
+        )
 
     if kwargs["tidal"]:
         client = TidalClient()
@@ -290,8 +320,10 @@ def config(ctx, **kwargs):
 
 
 def none_chosen():
+    """Print message if nothing was chosen."""
     click.secho("No items chosen, exiting.", fg="bright_red")
 
 
 def main():
+    """Run the main program."""
     cli(obj={})
