@@ -1,3 +1,5 @@
+"""The clients that interact with the service APIs."""
+
 import base64
 import hashlib
 import json
@@ -34,7 +36,7 @@ from .exceptions import (
 from .spoofbuz import Spoofer
 from .utils import gen_threadsafe_session, get_quality
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger("streamrip")
 
 
 class Client(ABC):
@@ -44,6 +46,10 @@ class Client(ABC):
     it is merely a template.
     """
 
+    source: str
+    max_quality: int
+    logged_in: bool
+
     @abstractmethod
     def login(self, **kwargs):
         """Authenticate the client.
@@ -72,35 +78,26 @@ class Client(ABC):
         pass
 
     @abstractmethod
-    def get_file_url(self, track_id, quality=3) -> Union[dict]:
+    def get_file_url(self, track_id, quality=3) -> Union[dict, str]:
         """Get the direct download url dict for a file.
 
         :param track_id: id of the track
         """
         pass
 
-    @property
-    @abstractmethod
-    def source(self):
-        """Source from which the Client retrieves data."""
-        pass
-
-    @property
-    @abstractmethod
-    def max_quality(self):
-        """The maximum quality that the Client supports."""
-        pass
-
 
 class QobuzClient(Client):
+    """QobuzClient."""
+
     source = "qobuz"
     max_quality = 4
 
     # ------- Public Methods -------------
     def __init__(self):
+        """Create a QobuzClient object."""
         self.logged_in = False
 
-    def login(self, email: str, pwd: str, **kwargs):
+    def login(self, **kwargs):
         """Authenticate the QobuzClient. Must have a paid membership.
 
         If `app_id` and `secrets` are not provided, this will run the
@@ -114,6 +111,8 @@ class QobuzClient(Client):
         :param kwargs: app_id: str, secrets: list, return_secrets: bool
         """
         click.secho(f"Logging into {self.source}", fg="green")
+        email: str = kwargs["email"]
+        pwd: str = kwargs["pwd"]
         if self.logged_in:
             logger.debug("Already logged in")
             return
@@ -140,6 +139,12 @@ class QobuzClient(Client):
         self.logged_in = True
 
     def get_tokens(self) -> Tuple[str, Sequence[str]]:
+        """Return app id and secrets.
+
+        These can be saved and reused.
+
+        :rtype: Tuple[str, Sequence[str]]
+        """
         return self.app_id, self.secrets
 
     def search(
@@ -178,16 +183,31 @@ class QobuzClient(Client):
         return self._api_search(query, media_type, limit)
 
     def get(self, item_id: Union[str, int], media_type: str = "album") -> dict:
-        return self._api_get(media_type, item_id=item_id)
+        """Get an item from the API.
+
+        :param item_id:
+        :type item_id: Union[str, int]
+        :param media_type:
+        :type media_type: str
+        :rtype: dict
+        """
+        resp = self._api_get(media_type, item_id=item_id)
+        logger.debug(resp)
+        return resp
 
     def get_file_url(self, item_id, quality=3) -> dict:
+        """Get the downloadble file url for a track.
+
+        :param item_id:
+        :param quality:
+        :rtype: dict
+        """
         return self._api_get_file_url(item_id, quality=quality)
 
     # ---------- Private Methods ---------------
 
-    def _gen_pages(self, epoint: str, params: dict) -> dict:
-        """When there are multiple pages of results, this lazily
-        yields them.
+    def _gen_pages(self, epoint: str, params: dict) -> Generator:
+        """When there are multiple pages of results, this yields them.
 
         :param epoint:
         :type epoint: str
@@ -196,7 +216,9 @@ class QobuzClient(Client):
         :rtype: dict
         """
         page, status_code = self._api_request(epoint, params)
-        logger.debug("Keys returned from _gen_pages: %s", ", ".join(page.keys()))
+        logger.debug(
+            "Keys returned from _gen_pages: %s", ", ".join(page.keys())
+        )
         key = epoint.split("/")[0] + "s"
         total = page.get(key, {})
         total = total.get("total") or total.get("items")
@@ -216,18 +238,19 @@ class QobuzClient(Client):
             yield page
 
     def _validate_secrets(self):
-        """Checks if the secrets are usable."""
+        """Check if the secrets are usable."""
         for secret in self.secrets:
             if self._test_secret(secret):
                 self.sec = secret
-                logger.debug("Working secret and app_id: %s - %s", secret, self.app_id)
+                logger.debug(
+                    "Working secret and app_id: %s - %s", secret, self.app_id
+                )
                 break
         if not hasattr(self, "sec"):
             raise InvalidAppSecretError(f"Invalid secrets: {self.secrets}")
 
     def _api_get(self, media_type: str, **kwargs) -> dict:
-        """Internal function that sends the request for metadata to the
-        Qobuz API.
+        """Request metadata from the Qobuz API.
 
         :param media_type:
         :type media_type: str
@@ -255,12 +278,16 @@ class QobuzClient(Client):
 
         response, status_code = self._api_request(epoint, params)
         if status_code != 200:
-            raise Exception(f'Error fetching metadata. "{response["message"]}"')
+            raise Exception(
+                f'Error fetching metadata. "{response["message"]}"'
+            )
 
         return response
 
-    def _api_search(self, query: str, media_type: str, limit: int = 500) -> Generator:
-        """Internal function that sends a search request to the API.
+    def _api_search(
+        self, query: str, media_type: str, limit: int = 500
+    ) -> Generator:
+        """Send a search request to the API.
 
         :param query:
         :type query: str
@@ -295,8 +322,7 @@ class QobuzClient(Client):
         return self._gen_pages(epoint, params)
 
     def _api_login(self, email: str, pwd: str):
-        """Internal function that logs into the api to get the user
-        authentication token.
+        """Log into the api to get the user authentication token.
 
         :param email:
         :type email: str
@@ -312,14 +338,18 @@ class QobuzClient(Client):
         resp, status_code = self._api_request(epoint, params)
 
         if status_code == 401:
-            raise AuthenticationError(f"Invalid credentials from params {params}")
+            raise AuthenticationError(
+                f"Invalid credentials from params {params}"
+            )
         elif status_code == 400:
             raise InvalidAppIdError(f"Invalid app id from params {params}")
         else:
             logger.info("Logged in to Qobuz")
 
         if not resp["user"]["credential"]["parameters"]:
-            raise IneligibleError("Free accounts are not eligible to download tracks.")
+            raise IneligibleError(
+                "Free accounts are not eligible to download tracks."
+            )
 
         self.uat = resp["user_auth_token"]
         self.session.headers.update({"X-User-Auth-Token": self.uat})
@@ -328,7 +358,7 @@ class QobuzClient(Client):
     def _api_get_file_url(
         self, track_id: Union[str, int], quality: int = 3, sec: str = None
     ) -> dict:
-        """Internal function that gets the file url given an id.
+        """Get the file url given a track id.
 
         :param track_id:
         :type track_id: Union[str, int]
@@ -353,7 +383,7 @@ class QobuzClient(Client):
         else:
             raise InvalidAppSecretError("Cannot find app secret")
 
-        quality = get_quality(quality, self.source)
+        quality = int(get_quality(quality, self.source))
         r_sig = f"trackgetFileUrlformat_id{quality}intentstreamtrack_id{track_id}{unix_ts}{secret}"
         logger.debug("Raw request signature: %s", r_sig)
         r_sig_hashed = hashlib.md5(r_sig.encode("utf-8")).hexdigest()
@@ -368,12 +398,14 @@ class QobuzClient(Client):
         }
         response, status_code = self._api_request("track/getFileUrl", params)
         if status_code == 400:
-            raise InvalidAppSecretError("Invalid app secret from params %s" % params)
+            raise InvalidAppSecretError(
+                "Invalid app secret from params %s" % params
+            )
 
         return response
 
     def _api_request(self, epoint: str, params: dict) -> Tuple[dict, int]:
-        """The function that handles all requests to the API.
+        """Send a request to the API.
 
         :param epoint:
         :type epoint: str
@@ -386,11 +418,13 @@ class QobuzClient(Client):
         try:
             return r.json(), r.status_code
         except Exception:
-            logger.error("Problem getting JSON. Status code: %s", r.status_code)
+            logger.error(
+                "Problem getting JSON. Status code: %s", r.status_code
+            )
             raise
 
     def _test_secret(self, secret: str) -> bool:
-        """Tests a secret.
+        """Test the authenticity of a secret.
 
         :param secret:
         :type secret: str
@@ -405,16 +439,21 @@ class QobuzClient(Client):
 
 
 class DeezerClient(Client):
+    """DeezerClient."""
+
     source = "deezer"
     max_quality = 2
 
     def __init__(self):
+        """Create a DeezerClient."""
         self.session = gen_threadsafe_session()
 
         # no login required
         self.logged_in = True
 
-    def search(self, query: str, media_type: str = "album", limit: int = 200) -> dict:
+    def search(
+        self, query: str, media_type: str = "album", limit: int = 200
+    ) -> dict:
         """Search API for query.
 
         :param query:
@@ -424,16 +463,20 @@ class DeezerClient(Client):
         :param limit:
         :type limit: int
         """
-        # TODO: more robust url sanitize
-        query = query.replace(" ", "+")
-
         # TODO: use limit parameter
-        response = self.session.get(f"{DEEZER_BASE}/search/{media_type}?q={query}")
+        response = self.session.get(
+            f"{DEEZER_BASE}/search/{media_type}", params={"q": query}
+        )
         response.raise_for_status()
-
         return response.json()
 
     def login(self, **kwargs):
+        """Return None.
+
+        Dummy method.
+
+        :param kwargs:
+        """
         logger.debug("Deezer does not require login call, returning")
 
     def get(self, meta_id: Union[str, int], media_type: str = "album"):
@@ -447,17 +490,27 @@ class DeezerClient(Client):
         url = f"{DEEZER_BASE}/{media_type}/{meta_id}"
         item = self.session.get(url).json()
         if media_type in ("album", "playlist"):
-            tracks = self.session.get(f"{url}/tracks").json()
+            tracks = self.session.get(
+                f"{url}/tracks", params={"limit": 1000}
+            ).json()
             item["tracks"] = tracks["data"]
             item["track_total"] = len(tracks["data"])
         elif media_type == "artist":
             albums = self.session.get(f"{url}/albums").json()
             item["albums"] = albums["data"]
 
+        logger.debug(item)
         return item
 
     @staticmethod
     def get_file_url(meta_id: Union[str, int], quality: int = 6):
+        """Get downloadable url for a track.
+
+        :param meta_id: The track ID.
+        :type meta_id: Union[str, int]
+        :param quality:
+        :type quality: int
+        """
         quality = min(DEEZER_MAX_Q, quality)
         url = f"{DEEZER_DL}/{get_quality(quality, 'deezer')}/{DEEZER_BASE}/track/{meta_id}"
         logger.debug(f"Download url {url}")
@@ -465,12 +518,15 @@ class DeezerClient(Client):
 
 
 class TidalClient(Client):
+    """TidalClient."""
+
     source = "tidal"
     max_quality = 3
 
     # ----------- Public Methods --------------
 
     def __init__(self):
+        """Create a TidalClient."""
         self.logged_in = False
 
         self.device_code = None
@@ -528,9 +584,13 @@ class TidalClient(Client):
         :param item_id:
         :param media_type:
         """
-        return self._api_get(item_id, media_type)
+        resp = self._api_get(item_id, media_type)
+        logger.debug(resp)
+        return resp
 
-    def search(self, query: str, media_type: str = "album", limit: int = 100) -> dict:
+    def search(
+        self, query: str, media_type: str = "album", limit: int = 100
+    ) -> dict:
         """Search for a query.
 
         :param query:
@@ -559,13 +619,19 @@ class TidalClient(Client):
             return self._get_video_stream_url(track_id)
 
         params = {
-            "audioquality": get_quality(min(quality, TIDAL_MAX_Q), self.source),
+            "audioquality": get_quality(
+                min(quality, TIDAL_MAX_Q), self.source
+            ),
             "playbackmode": "STREAM",
             "assetpresentation": "FULL",
         }
-        resp = self._api_request(f"tracks/{track_id}/playbackinfopostpaywall", params)
+        resp = self._api_request(
+            f"tracks/{track_id}/playbackinfopostpaywall", params
+        )
         try:
-            manifest = json.loads(base64.b64decode(resp["manifest"]).decode("utf-8"))
+            manifest = json.loads(
+                base64.b64decode(resp["manifest"]).decode("utf-8")
+            )
         except KeyError:
             raise Exception(resp["userMessage"])
 
@@ -577,7 +643,7 @@ class TidalClient(Client):
         }
 
     def get_tokens(self) -> dict:
-        """Used for saving them for later use.
+        """Return tokens to save for later use.
 
         :rtype: dict
         """
@@ -594,21 +660,23 @@ class TidalClient(Client):
 
     # ------------ Utilities to login -------------
 
-    def _login_new_user(self, launch=True):
-        """This will launch the browser and ask the user to log into tidal.
+    def _login_new_user(self, launch: bool = True):
+        """Create app url where the user can log in.
 
-        :param launch:
+        :param launch: Launch the browser.
+        :type launch: bool
         """
         login_link = f"https://{self._get_device_code()}"
 
         click.secho(
-            f"Go to {login_link} to log into Tidal within 5 minutes.", fg="blue"
+            f"Go to {login_link} to log into Tidal within 5 minutes.",
+            fg="blue",
         )
         if launch:
             click.launch(login_link)
 
         start = time.time()
-        elapsed = 0
+        elapsed = 0.0
         while elapsed < 600:  # 5 mins to login
             elapsed = time.time() - start
             status = self._check_auth_status()
@@ -689,7 +757,9 @@ class TidalClient(Client):
         return True
 
     def _refresh_access_token(self):
-        """The access token expires in a week, so it must be refreshed.
+        """Refresh the access token given a refresh token.
+
+        The access token expires in a week, so it must be refreshed.
         Requires a refresh token.
         """
         data = {
@@ -714,7 +784,9 @@ class TidalClient(Client):
         self._update_authorization()
 
     def _login_by_access_token(self, token, user_id=None):
-        """This is the method used to login after the access token has been saved.
+        """Login using the access token.
+
+        Used after the initial authorization.
 
         :param token:
         :param user_id: Not necessary.
@@ -740,7 +812,7 @@ class TidalClient(Client):
 
     @property
     def authorization(self):
-        """The auth header."""
+        """Get the auth header."""
         return {"authorization": f"Bearer {self.access_token}"}
 
     # ------------- Fetch data ------------------
@@ -765,7 +837,9 @@ class TidalClient(Client):
                     offset += 100
                     tracks_left -= 100
                     resp["items"].extend(
-                        self._api_request(f"{url}/items", {"offset": offset})["items"]
+                        self._api_request(f"{url}/items", {"offset": offset})[
+                            "items"
+                        ]
                     )
 
             item["tracks"] = [item["item"] for item in resp["items"]]
@@ -776,7 +850,7 @@ class TidalClient(Client):
         return item
 
     def _api_request(self, path: str, params=None) -> dict:
-        """The function that handles all tidal API requests.
+        """Handle Tidal API requests.
 
         :param path:
         :type path: str
@@ -792,8 +866,7 @@ class TidalClient(Client):
         return r
 
     def _get_video_stream_url(self, video_id: str) -> str:
-        """Videos have to be ripped from an hls stream, so they require
-        seperate processing.
+        """Get the HLS video stream url.
 
         :param video_id:
         :type video_id: str
@@ -811,7 +884,9 @@ class TidalClient(Client):
             r'#EXT-X-STREAM-INF:BANDWIDTH=\d+,AVERAGE-BANDWIDTH=\d+,CODECS="[^"]+"'
             r",RESOLUTION=\d+x\d+\n(.+)"
         )
-        manifest = json.loads(base64.b64decode(resp["manifest"]).decode("utf-8"))
+        manifest = json.loads(
+            base64.b64decode(resp["manifest"]).decode("utf-8")
+        )
         available_urls = self.session.get(manifest["urls"][0])
         url_info = re.findall(stream_url_regex, available_urls.text)
 
@@ -819,7 +894,7 @@ class TidalClient(Client):
         return url_info[-1]
 
     def _api_post(self, url, data, auth=None):
-        """Function used for posting to tidal API.
+        """Post to the Tidal API.
 
         :param url:
         :param data:
@@ -830,11 +905,14 @@ class TidalClient(Client):
 
 
 class SoundCloudClient(Client):
+    """SoundCloudClient."""
+
     source = "soundcloud"
     max_quality = 0
     logged_in = True
 
     def __init__(self):
+        """Create a SoundCloudClient."""
         self.session = gen_threadsafe_session(headers={"User-Agent": AGENT})
 
     def login(self):
@@ -847,7 +925,10 @@ class SoundCloudClient(Client):
         :param id:
         :param media_type:
         """
-        assert media_type in ("track", "playlist"), f"{media_type} not supported"
+        assert media_type in (
+            "track",
+            "playlist",
+        ), f"{media_type} not supported"
 
         if "http" in str(id):
             resp, _ = self._get(f"resolve?url={id}")
@@ -856,9 +937,10 @@ class SoundCloudClient(Client):
         else:
             raise Exception(id)
 
+        logger.debug(resp)
         return resp
 
-    def get_file_url(self, track: dict, quality) -> dict:
+    def get_file_url(self, track, quality):
         """Get the streamable file url from soundcloud.
 
         It will most likely be an hls stream, which will have to be manually
@@ -869,6 +951,9 @@ class SoundCloudClient(Client):
         :param quality:
         :rtype: dict
         """
+        # TODO: find better solution for typing
+        assert isinstance(track, dict)
+
         if not track["streamable"] or track["policy"] == "BLOCK":
             raise Exception
 
@@ -880,7 +965,10 @@ class SoundCloudClient(Client):
             url = None
             for tc in track["media"]["transcodings"]:
                 fmt = tc["format"]
-                if fmt["protocol"] == "hls" and fmt["mime_type"] == "audio/mpeg":
+                if (
+                    fmt["protocol"] == "hls"
+                    and fmt["mime_type"] == "audio/mpeg"
+                ):
                     url = tc["url"]
                     break
 
@@ -902,8 +990,7 @@ class SoundCloudClient(Client):
         return resp
 
     def _get(self, path, params=None, no_base=False, resp_obj=False):
-        """The lower level of `SoundCloudClient.get` that handles request
-        parameters and other options.
+        """Send a request to the SoundCloud API.
 
         :param path:
         :param params:
