@@ -1,3 +1,5 @@
+"""Constants that are kept in one place."""
+
 import os
 from pathlib import Path
 
@@ -12,11 +14,14 @@ CONFIG_PATH = os.path.join(CONFIG_DIR, "config.yaml")
 LOG_DIR = click.get_app_dir(APPNAME)
 DB_PATH = os.path.join(LOG_DIR, "downloads.db")
 
-DOWNLOADS_DIR = os.path.join(Path.home(), "StreamripDownloads")
+HOME = Path.home()
+DOWNLOADS_DIR = os.path.join(HOME, "StreamripDownloads")
 
 AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:83.0) Gecko/20100101 Firefox/83.0"
 
-TIDAL_COVER_URL = "https://resources.tidal.com/images/{uuid}/{width}x{height}.jpg"
+TIDAL_COVER_URL = (
+    "https://resources.tidal.com/images/{uuid}/{width}x{height}.jpg"
+)
 
 
 QUALITY_DESC = {
@@ -67,6 +72,7 @@ __MP4_KEYS = (
     "disk",
     None,
     None,
+    None,
 )
 
 __MP3_KEYS = (
@@ -90,6 +96,7 @@ __MP3_KEYS = (
     id3.TPOS,
     None,
     None,
+    None,
 )
 
 __METADATA_TYPES = (
@@ -113,6 +120,7 @@ __METADATA_TYPES = (
     "discnumber",
     "tracktotal",
     "disctotal",
+    "date",
 )
 
 
@@ -126,11 +134,17 @@ FLAC_MAX_BLOCKSIZE = 16777215  # 16.7 MB
 
 # TODO: give these more descriptive names
 TRACK_KEYS = ("tracknumber", "artist", "albumartist", "composer", "title")
-ALBUM_KEYS = ("albumartist", "title", "year", "bit_depth", "sampling_rate", "container")
-# TODO: rename these to DEFAULT_FOLDER_FORMAT etc
-FOLDER_FORMAT = (
-    "{albumartist} - {title} ({year}) [{container}] [{bit_depth}B-{sampling_rate}kHz]"
+ALBUM_KEYS = (
+    "albumartist",
+    "title",
+    "year",
+    "bit_depth",
+    "sampling_rate",
+    "container",
+    "composer",
 )
+# TODO: rename these to DEFAULT_FOLDER_FORMAT etc
+FOLDER_FORMAT = "{albumartist} - {title} ({year}) [{container}] [{bit_depth}B-{sampling_rate}kHz]"
 TRACK_FORMAT = "{tracknumber}. {artist} - {title}"
 
 
@@ -145,7 +159,7 @@ LASTFM_URL_REGEX = r"https://www.last.fm/user/\w+/playlists/\w+"
 QOBUZ_INTERPRETER_URL_REGEX = (
     r"https?://www\.qobuz\.com/\w\w-\w\w/interpreter/[-\w]+/[-\w]+"
 )
-
+YOUTUBE_URL_REGEX = r"https://www\.youtube\.com/watch\?v=[-\w]+"
 
 TIDAL_MAX_Q = 7
 
