@@ -1,3 +1,5 @@
+"""Wrapper classes over FFMPEG."""
+
 import logging
 import os
 import shutil
@@ -7,7 +9,7 @@ from typing import Optional
 
 from .exceptions import ConversionError
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger("streamrip")
 
 SAMPLING_RATES = (44100, 48000, 88200, 96000, 176400, 192000)
 
@@ -15,11 +17,11 @@ SAMPLING_RATES = (44100, 48000, 88200, 96000, 176400, 192000)
 class Converter:
     """Base class for audio codecs."""
 
-    codec_name = None
-    codec_lib = None
-    container = None
-    lossless = False
-    default_ffmpeg_arg = ""
+    codec_name: str
+    codec_lib: str
+    container: str
+    lossless: bool = False
+    default_ffmpeg_arg: str = ""
 
     def __init__(
         self,
@@ -31,7 +33,8 @@ class Converter:
         remove_source: bool = False,
         show_progress: bool = False,
     ):
-        """
+        """Create a Converter object.
+
         :param filename:
         :type filename: str
         :param ffmpeg_arg: The codec ffmpeg argument (defaults to an "optimal value")
@@ -42,14 +45,16 @@ class Converter:
         :type bit_depth: Optional[int]
         :param copy_art: Embed the cover art (if found) into the encoded file
         :type copy_art: bool
-        :param remove_source:
+        :param remove_source: Remove the source file after conversion.
         :type remove_source: bool
         """
         logger.debug(locals())
 
         self.filename = filename
         self.final_fn = f"{os.path.splitext(filename)[0]}.{self.container}"
-        self.tempfile = os.path.join(gettempdir(), os.path.basename(self.final_fn))
+        self.tempfile = os.path.join(
+            gettempdir(), os.path.basename(self.final_fn)
+        )
         self.remove_source = remove_source
         self.sampling_rate = sampling_rate
         self.bit_depth = bit_depth
@@ -112,9 +117,13 @@ class Converter:
         if self.lossless:
             if isinstance(self.sampling_rate, int):
                 sampling_rates = "|".join(
-                    str(rate) for rate in SAMPLING_RATES if rate <= self.sampling_rate
+                    str(rate)
+                    for rate in SAMPLING_RATES
+                    if rate <= self.sampling_rate
+                )
+                command.extend(
+                    ["-af", f"aformat=sample_rates={sampling_rates}"]
                 )
-                command.extend(["-af", f"aformat=sample_rates={sampling_rates}"])
 
             elif self.sampling_rate is not None:
                 raise TypeError(
@@ -129,7 +138,9 @@ class Converter:
                 else:
                     raise ValueError("Bit depth must be 16, 24, or 32")
             elif self.bit_depth is not None:
-                raise TypeError(f"Bit depth must be int, not {type(self.bit_depth)}")
+                raise TypeError(
+                    f"Bit depth must be int, not {type(self.bit_depth)}"
+                )
 
         # automatically overwrite
         command.extend(["-y", self.tempfile])
@@ -148,7 +159,8 @@ class Converter:
 
 
 class FLAC(Converter):
-    " Class for FLAC converter. "
+    """Class for FLAC converter."""
+
     codec_name = "flac"
     codec_lib = "flac"
     container = "flac"
@@ -156,8 +168,9 @@ class FLAC(Converter):
 
 
 class LAME(Converter):
-    """
-    Class for libmp3lame converter. Defaul ffmpeg_arg: `-q:a 0`.
+    """Class for libmp3lame converter.
+
+    Default ffmpeg_arg: `-q:a 0`.
 
     See available options:
     https://trac.ffmpeg.org/wiki/Encode/MP3
@@ -170,7 +183,8 @@ class LAME(Converter):
 
 
 class ALAC(Converter):
-    " Class for ALAC converter. "
+    """Class for ALAC converter."""
+
     codec_name = "alac"
     codec_lib = "alac"
     container = "m4a"
@@ -178,8 +192,9 @@ class ALAC(Converter):
 
 
 class Vorbis(Converter):
-    """
-    Class for libvorbis converter. Default ffmpeg_arg: `-q:a 6`.
+    """Class for libvorbis converter.
+
+    Default ffmpeg_arg: `-q:a 6`.
 
     See available options:
     https://trac.ffmpeg.org/wiki/TheoraVorbisEncodingGuide
@@ -188,12 +203,15 @@ class Vorbis(Converter):
     codec_name = "vorbis"
     codec_lib = "libvorbis"
     container = "ogg"
-    default_ffmpeg_arg = "-q:a 6"  # 160, aka the "high" quality profile from Spotify
+    default_ffmpeg_arg = (
+        "-q:a 6"  # 160, aka the "high" quality profile from Spotify
+    )
 
 
 class OPUS(Converter):
-    """
-    Class for libopus. Default ffmpeg_arg: `-b:a 128 -vbr on`.
+    """Class for libopus.
+
+    Default ffmpeg_arg: `-b:a 128 -vbr on`.
 
     See more:
     http://ffmpeg.org/ffmpeg-codecs.html#libopus-1
@@ -206,8 +224,9 @@ class OPUS(Converter):
 
 
 class AAC(Converter):
-    """
-    Class for libfdk_aac converter. Default ffmpeg_arg: `-b:a 256k`.
+    """Class for libfdk_aac converter.
+
+    Default ffmpeg_arg: `-b:a 256k`.
 
     See available options:
     https://trac.ffmpeg.org/wiki/Encode/AAC
