@@ -1,4 +1,7 @@
+"""The stuff that ties everything together for the CLI to use."""
+
 import concurrent.futures
+import html
 import logging
 import os
 import re
@@ -6,14 +9,20 @@ import sys
 from getpass import getpass
 from hashlib import md5
 from string import Formatter
-from typing import Generator, Optional, Tuple, Union
+from typing import Dict, Generator, List, Optional, Tuple, Type, Union
 
 import click
 import requests
 from tqdm import tqdm
 
-from .bases import Track, Video
-from .clients import DeezerClient, QobuzClient, SoundCloudClient, TidalClient
+from .bases import Track, Video, YoutubeVideo
+from .clients import (
+    Client,
+    DeezerClient,
+    QobuzClient,
+    SoundCloudClient,
+    TidalClient,
+)
 from .config import Config
 from .constants import (
     CONFIG_PATH,
@@ -23,6 +32,7 @@ from .constants import (
     QOBUZ_INTERPRETER_URL_REGEX,
     SOUNDCLOUD_URL_REGEX,
     URL_REGEX,
+    YOUTUBE_URL_REGEX,
 )
 from .db import MusicDB
 from .exceptions import (
@@ -34,10 +44,18 @@ from .exceptions import (
 from .tracklists import Album, Artist, Label, Playlist, Tracklist
 from .utils import extract_interpreter_url
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger("streamrip")
 
 
-MEDIA_CLASS = {
+Media = Union[
+    Type[Album],
+    Type[Playlist],
+    Type[Artist],
+    Type[Track],
+    Type[Label],
+    Type[Video],
+]
+MEDIA_CLASS: Dict[str, Media] = {
     "album": Album,
     "playlist": Playlist,
     "artist": Artist,
@@ -45,23 +63,31 @@ MEDIA_CLASS = {
     "label": Label,
     "video": Video,
 }
-Media = Union[Album, Playlist, Artist, Track]
 
 
 class MusicDL(list):
+    """MusicDL."""
+
     def __init__(
         self,
         config: Optional[Config] = None,
     ):
+        """Create a MusicDL object.
 
+        :param config:
+        :type config: Optional[Config]
+        """
         self.url_parse = re.compile(URL_REGEX)
         self.soundcloud_url_parse = re.compile(SOUNDCLOUD_URL_REGEX)
         self.lastfm_url_parse = re.compile(LASTFM_URL_REGEX)
         self.interpreter_url_parse = re.compile(QOBUZ_INTERPRETER_URL_REGEX)
+        self.youtube_url_parse = re.compile(YOUTUBE_URL_REGEX)
 
-        self.config = config
-        if self.config is None:
+        self.config: Config
+        if config is None:
             self.config = Config(CONFIG_PATH)
+        else:
+            self.config = config
 
         self.clients = {
             "qobuz": QobuzClient(),
@@ -70,26 +96,44 @@ class MusicDL(list):
             "soundcloud": SoundCloudClient(),
         }
 
-        if config.session["database"]["enabled"]:
-            if config.session["database"]["path"] is not None:
-                self.db = MusicDB(config.session["database"]["path"])
+        self.db: Union[MusicDB, list]
+        if self.config.session["database"]["enabled"]:
+            if self.config.session["database"]["path"] is not None:
+                self.db = MusicDB(self.config.session["database"]["path"])
             else:
                 self.db = MusicDB(DB_PATH)
-                config.file["database"]["path"] = DB_PATH
-                config.save()
+                self.config.file["database"]["path"] = DB_PATH
+                self.config.save()
         else:
             self.db = []
 
     def handle_urls(self, url: str):
-        """Download a url
+        """Download a url.
 
         :param url:
         :type url: str
         :raises InvalidSourceError
         :raises ParsingError
         """
+        # youtube is handled by youtube-dl, so much of the
+        # processing is not necessary
+        youtube_urls = self.youtube_url_parse.findall(url)
+        if youtube_urls != []:
+            self.extend(YoutubeVideo(u) for u in youtube_urls)
 
-        for source, url_type, item_id in self.parse_urls(url):
+        parsed = self.parse_urls(url)
+        if not parsed and len(self) == 0:
+            if "last.fm" in url:
+                message = (
+                    f"For last.fm urls, use the {click.style('lastfm', fg='yellow')} "
+                    f"command. See {click.style('rip lastfm --help', fg='yellow')}."
+                )
+            else:
+                message = url
+
+            raise ParsingError(message)
+
+        for source, url_type, item_id in parsed:
             if item_id in self.db:
                 logger.info(
                     f"ID {item_id} already downloaded, use --no-db to override."
@@ -103,6 +147,15 @@ class MusicDL(list):
             self.handle_item(source, url_type, item_id)
 
     def handle_item(self, source: str, media_type: str, item_id: str):
+        """Get info and parse into a Media object.
+
+        :param source:
+        :type source: str
+        :param media_type:
+        :type media_type: str
+        :param item_id:
+        :type item_id: str
+        """
         self.assert_creds(source)
 
         client = self.get_client(source)
@@ -116,6 +169,10 @@ class MusicDL(list):
         self.append(item)
 
     def _get_download_args(self) -> dict:
+        """Get the arguments to pass to Media.download.
+
+        :rtype: dict
+        """
         return {
             "database": self.db,
             "parent_folder": self.config.session["downloads"]["folder"],
@@ -123,66 +180,79 @@ class MusicDL(list):
             "track_format": self.config.session["path_format"]["track"],
             "embed_cover": self.config.session["artwork"]["embed"],
             "embed_cover_size": self.config.session["artwork"]["size"],
-            "keep_hires_cover": self.config.session["artwork"]["keep_hires_cover"],
+            "keep_hires_cover": self.config.session["artwork"][
+                "keep_hires_cover"
+            ],
             "set_playlist_to_album": self.config.session["metadata"][
                 "set_playlist_to_album"
             ],
             "stay_temp": self.config.session["conversion"]["enabled"],
             "conversion": self.config.session["conversion"],
-            "concurrent_downloads": self.config.session["concurrent_downloads"],
+            "concurrent_downloads": self.config.session[
+                "concurrent_downloads"
+            ],
             "new_tracknumbers": self.config.session["metadata"][
                 "new_playlist_tracknumbers"
             ],
             "download_videos": self.config.session["tidal"]["download_videos"],
-            "download_booklets": self.config.session["qobuz"]["download_booklets"],
+            "download_booklets": self.config.session["qobuz"][
+                "download_booklets"
+            ],
+            "download_youtube_videos": self.config.session["youtube"][
+                "download_videos"
+            ],
+            "youtube_video_downloads_folder": self.config.session["youtube"][
+                "video_downloads_folder"
+            ],
         }
 
     def download(self):
+        """Download all the items in self."""
         try:
             arguments = self._get_download_args()
         except KeyError:
-            click.secho(
-                "Updating config file... Some settings may be lost. Please run the "
-                "command again.",
-                fg="magenta",
-            )
+            self._config_updating_message()
             self.config.update()
             exit()
         except Exception as err:
-            click.secho(
-                "There was a problem with your config file. This happens "
-                "sometimes after updates. Run ",
-                nl=False,
-                fg="red",
-            )
-            click.secho("rip config --reset ", fg="yellow", nl=False)
-            click.secho("to reset it. You will need to log in again.", fg="red")
-            logger.debug(err)
+            self._config_corrupted_message(err)
             exit()
 
         logger.debug("Arguments from config: %s", arguments)
 
-        source_subdirs = self.config.session["downloads"]["source_subdirectories"]
+        source_subdirs = self.config.session["downloads"][
+            "source_subdirectories"
+        ]
         for item in self:
-
             if source_subdirs:
                 arguments["parent_folder"] = self.__get_source_subdir(
                     item.client.source
                 )
 
-            arguments["quality"] = self.config.session[item.client.source]["quality"]
+            if item is YoutubeVideo:
+                item.download(**arguments)
+                continue
+
+            arguments["quality"] = self.config.session[item.client.source][
+                "quality"
+            ]
             if isinstance(item, Artist):
                 filters_ = tuple(
                     k for k, v in self.config.session["filters"].items() if v
                 )
                 arguments["filters"] = filters_
-                logger.debug("Added filter argument for artist/label: %s", filters_)
+                logger.debug(
+                    "Added filter argument for artist/label: %s", filters_
+                )
 
             if not (isinstance(item, Tracklist) and item.loaded):
+                logger.debug("Loading metadata")
                 try:
                     item.load_meta()
                 except NonStreamable:
-                    click.secho(f"{item!s} is not available, skipping.", fg="red")
+                    click.secho(
+                        f"{item!s} is not available, skipping.", fg="red"
+                    )
                     continue
 
             item.download(**arguments)
@@ -194,7 +264,13 @@ class MusicDL(list):
             if self.db != [] and hasattr(item, "id"):
                 self.db.add(item.id)
 
-    def get_client(self, source: str):
+    def get_client(self, source: str) -> Client:
+        """Get a client given the source and log in.
+
+        :param source:
+        :type source: str
+        :rtype: Client
+        """
         client = self.clients[source]
         if not client.logged_in:
             self.assert_creds(source)
@@ -202,6 +278,10 @@ class MusicDL(list):
         return client
 
     def login(self, client):
+        """Log into a client, if applicable.
+
+        :param client:
+        """
         creds = self.config.creds(client.source)
         if not client.logged_in:
             while True:
@@ -225,8 +305,8 @@ class MusicDL(list):
                 self.config.file["tidal"].update(client.get_tokens())
                 self.config.save()
 
-    def parse_urls(self, url: str) -> Tuple[str, str]:
-        """Returns the type of the url and the id.
+    def parse_urls(self, url: str) -> List[Tuple[str, str, str]]:
+        """Return the type of the url and the id.
 
         Compatible with urls of the form:
             https://www.qobuz.com/us-en/{type}/{name}/{id}
@@ -239,8 +319,7 @@ class MusicDL(list):
 
         :raises exceptions.ParsingError
         """
-
-        parsed = []
+        parsed: List[Tuple[str, str, str]] = []
 
         interpreter_urls = self.interpreter_url_parse.findall(url)
         if interpreter_urls:
@@ -257,7 +336,9 @@ class MusicDL(list):
 
         parsed.extend(self.url_parse.findall(url))  # Qobuz, Tidal, Dezer
         soundcloud_urls = self.soundcloud_url_parse.findall(url)
-        soundcloud_items = [self.clients["soundcloud"].get(u) for u in soundcloud_urls]
+        soundcloud_items = [
+            self.clients["soundcloud"].get(u) for u in soundcloud_urls
+        ]
 
         parsed.extend(
             ("soundcloud", item["kind"], url)
@@ -266,54 +347,113 @@ class MusicDL(list):
 
         logger.debug(f"Parsed urls: {parsed}")
 
-        if parsed != []:
-            return parsed
+        return parsed
 
-        raise ParsingError(f"Error parsing URL: `{url}`")
+    def handle_lastfm_urls(self, urls: str):
+        """Get info from lastfm url, and parse into Media objects.
 
-    def handle_lastfm_urls(self, urls):
+        This works by scraping the last.fm page and using a regex to
+        find the track titles and artists. The information is queried
+        in a Client.search(query, 'track') call and the first result is
+        used.
+
+        :param urls:
+        """
+
+        # Available keys: ['artist', 'title']
+        QUERY_FORMAT: Dict[str, str] = {
+            "tidal": "{title}",
+            "qobuz": "{title} {artist}",
+            "deezer": "{title} {artist}",
+            "soundcloud": "{title} {artist}",
+        }
+
+        # For testing:
         # https://www.last.fm/user/nathan3895/playlists/12058911
-        user_regex = re.compile(r"https://www\.last\.fm/user/([^/]+)/playlists/\d+")
+        user_regex = re.compile(
+            r"https://www\.last\.fm/user/([^/]+)/playlists/\d+"
+        )
         lastfm_urls = self.lastfm_url_parse.findall(urls)
-        lastfm_source = self.config.session["lastfm"]["source"]
-        tracks_not_found = 0
+        try:
+            lastfm_source = self.config.session["lastfm"]["source"]
+            lastfm_fallback_source = self.config.session["lastfm"][
+                "fallback_source"
+            ]
+        except KeyError:
+            self._config_updating_message()
+            self.config.update()
+            exit()
+        except Exception as err:
+            self._config_corrupted_message(err)
+            raise click.Abort
 
-        def search_query(query: str, playlist: Playlist):
-            global tracks_not_found
-            try:
-                track = next(self.search(lastfm_source, query, media_type="track"))
-                if self.config.session["metadata"]["set_playlist_to_album"]:
-                    # so that the playlist name (actually the album) isn't
-                    # amended to include version and work tags from individual tracks
-                    track.meta.version = track.meta.work = None
+        def search_query(title, artist, playlist) -> bool:
+            """Search for a query and add the first result to playlist.
 
-                playlist.append(track)
-            except NoResultsFound:
-                tracks_not_found += 1
-                return
+            :param query:
+            :type query: str
+            :param playlist:
+            :type playlist: Playlist
+            :rtype: bool
+            """
+
+            def try_search(source) -> Optional[Track]:
+                if source == lastfm_fallback_source:
+                    click.secho("using fallback", fg="red")
+                try:
+                    query = QUERY_FORMAT[lastfm_source].format(
+                        title=title, artist=artist
+                    )
+                    return next(self.search(source, query, media_type="track"))
+                except (NoResultsFound, StopIteration):
+                    return None
+
+            track = try_search(lastfm_source) or try_search(
+                lastfm_fallback_source
+            )
+            if track is None:
+                return False
+
+            if self.config.session["metadata"]["set_playlist_to_album"]:
+                # so that the playlist name (actually the album) isn't
+                # amended to include version and work tags from individual tracks
+                track.meta.version = track.meta.work = None
+
+            playlist.append(track)
+            return True
 
         for purl in lastfm_urls:
             click.secho(f"Fetching playlist at {purl}", fg="blue")
             title, queries = self.get_lastfm_playlist(purl)
 
             pl = Playlist(client=self.get_client(lastfm_source), name=title)
-            pl.creator = user_regex.search(purl).group(1)
+            creator_match = user_regex.search(purl)
+            if creator_match is not None:
+                pl.creator = creator_match.group(1)
 
-            with concurrent.futures.ThreadPoolExecutor(max_workers=15) as executor:
+            tracks_not_found = 0
+            with concurrent.futures.ThreadPoolExecutor(
+                max_workers=15
+            ) as executor:
                 futures = [
-                    executor.submit(search_query, f"{title} {artist}", pl)
+                    executor.submit(search_query, title, artist, pl)
                     for title, artist in queries
                 ]
                 # only for the progress bar
-                for f in tqdm(
+                for search_attempt in tqdm(
                     concurrent.futures.as_completed(futures),
                     total=len(futures),
                     desc="Searching",
                 ):
-                    pass
+                    if not search_attempt.result():
+                        tracks_not_found += 1
 
             pl.loaded = True
-            click.secho(f"{tracks_not_found} tracks not found.", fg="yellow")
+
+            if tracks_not_found > 0:
+                click.secho(
+                    f"{tracks_not_found} tracks not found.", fg="yellow"
+                )
             self.append(pl)
 
     def handle_txt(self, filepath: Union[str, os.PathLike]):
@@ -329,12 +469,29 @@ class MusicDL(list):
             self.handle_urls(txt.read())
 
     def search(
-        self, source: str, query: str, media_type: str = "album", limit: int = 200
+        self,
+        source: str,
+        query: str,
+        media_type: str = "album",
+        limit: int = 200,
     ) -> Generator:
+        """Universal search.
+
+        :param source:
+        :type source: str
+        :param query:
+        :type query: str
+        :param media_type:
+        :type media_type: str
+        :param limit: Not Implemented
+        :type limit: int
+        :rtype: Generator
+        """
+        logger.debug("searching for %s", query)
+
         client = self.get_client(source)
         results = client.search(query, media_type)
 
-        i = 0
         if isinstance(results, Generator):  # QobuzClient
             for page in results:
                 tracklist = (
@@ -342,27 +499,34 @@ class MusicDL(list):
                     if media_type != "featured"
                     else page["albums"]["items"]
                 )
-                for item in tracklist:
-                    yield MEDIA_CLASS[
+                for i, item in enumerate(tracklist):
+                    yield MEDIA_CLASS[  # type: ignore
                         media_type if media_type != "featured" else "album"
                     ].from_api(item, client)
-                    i += 1
                     if i > limit:
                         return
         else:
+            logger.debug("Not generator")
             items = (
-                results.get("data") or results.get("items") or results.get("collection")
+                results.get("data")
+                or results.get("items")
+                or results.get("collection")
             )
             if items is None:
                 raise NoResultsFound(query)
+            logger.debug("Number of results: %d", len(items))
 
-            for item in items:
-                yield MEDIA_CLASS[media_type].from_api(item, client)
-                i += 1
+            for i, item in enumerate(items):
+                logger.debug(item["title"])
+                yield MEDIA_CLASS[media_type].from_api(item, client)  # type: ignore
                 if i > limit:
                     return
 
-    def preview_media(self, media):
+    def preview_media(self, media) -> str:
+        """Return a preview string of a Media object.
+
+        :param media:
+        """
         if isinstance(media, Album):
             fmt = (
                 "{albumartist} - {album}\n"
@@ -386,12 +550,23 @@ class MusicDL(list):
             raise NotImplementedError
 
         fields = (fname for _, fname, _, _ in Formatter().parse(fmt) if fname)
-        ret = fmt.format(**{k: media.get(k, default="Unknown") for k in fields})
+        ret = fmt.format(
+            **{k: media.get(k, default="Unknown") for k in fields}
+        )
         return ret
 
-    def interactive_search(
+    def interactive_search(  # noqa
         self, query: str, source: str = "qobuz", media_type: str = "album"
     ):
+        """Show an interactive menu that contains search results.
+
+        :param query:
+        :type query: str
+        :param source:
+        :type source: str
+        :param media_type:
+        :type media_type: str
+        """
         results = tuple(self.search(source, query, media_type, limit=50))
 
         def title(res):
@@ -472,6 +647,18 @@ class MusicDL(list):
                 return True
 
     def get_lastfm_playlist(self, url: str) -> Tuple[str, list]:
+        """From a last.fm url, find the playlist title and tracks.
+
+        Each page contains 50 results, so `num_tracks // 50 + 1` requests
+        are sent per playlist.
+
+        :param url:
+        :type url: str
+        :rtype: Tuple[str, list]
+        """
+
+        logger.debug("Fetching lastfm playlist")
+
         info = []
         words = re.compile(r"[\w\s]+")
         title_tags = re.compile('title="([^"]+)"')
@@ -487,20 +674,44 @@ class MusicDL(list):
 
         r = requests.get(url)
         get_titles(r.text)
-        remaining_tracks = (
-            int(re.search(r'data-playlisting-entry-count="(\d+)"', r.text).group(1))
-            - 50
+        remaining_tracks_match = re.search(
+            r'data-playlisting-entry-count="(\d+)"', r.text
         )
-        playlist_title = re.search(
-            r'<h1 class="playlisting-playlist-header-title">([^<]+)</h1>', r.text
-        ).group(1)
+        if remaining_tracks_match is None:
+            raise ParsingError("Error parsing lastfm page: %s", r.text)
 
-        page = 1
-        while remaining_tracks > 0:
-            page += 1
-            r = requests.get(f"{url}?page={page}")
-            get_titles(r.text)
-            remaining_tracks -= 50
+        total_tracks = int(remaining_tracks_match.group(1))
+        logger.debug("Total tracks: %d", total_tracks)
+        remaining_tracks = total_tracks - 50
+
+        playlist_title_match = re.search(
+            r'<h1 class="playlisting-playlist-header-title">([^<]+)</h1>',
+            r.text,
+        )
+        if playlist_title_match is None:
+            raise ParsingError("Error finding title from response")
+
+        playlist_title = html.unescape(playlist_title_match.group(1))
+
+        if remaining_tracks > 0:
+            with concurrent.futures.ThreadPoolExecutor(
+                max_workers=15
+            ) as executor:
+                last_page = int(remaining_tracks // 50) + int(
+                    remaining_tracks % 50 != 0
+                )
+
+                futures = [
+                    executor.submit(requests.get, f"{url}?page={page}")
+                    for page in range(1, last_page + 1)
+                ]
+
+            for future in tqdm(
+                concurrent.futures.as_completed(futures),
+                total=len(futures),
+                desc="Scraping playlist",
+            ):
+                get_titles(future.result().text)
 
         return playlist_title, info
 
@@ -526,11 +737,18 @@ class MusicDL(list):
             ).hexdigest()
 
             self.config.save()
-            click.secho(f'Credentials saved to config file at "{self.config._path}"')
+            click.secho(
+                f'Credentials saved to config file at "{self.config._path}"'
+            )
         else:
             raise Exception
 
     def assert_creds(self, source: str):
+        """Ensure that the credentials for `source` are valid.
+
+        :param source:
+        :type source: str
+        """
         assert source in (
             "qobuz",
             "tidal",
@@ -549,3 +767,21 @@ class MusicDL(list):
             or self.config.file[source]["password"] is None
         ):
             self.prompt_creds(source)
+
+    def _config_updating_message(self):
+        click.secho(
+            "Updating config file... Some settings may be lost. Please run the "
+            "command again.",
+            fg="magenta",
+        )
+
+    def _config_corrupted_message(self, err: Exception):
+        click.secho(
+            "There was a problem with your config file. This happens "
+            "sometimes after updates. Run ",
+            nl=False,
+            fg="red",
+        )
+        click.secho("rip config --reset ", fg="yellow", nl=False)
+        click.secho("to reset it. You will need to log in again.", fg="red")
+        click.secho(str(err), fg="red")
