@@ -1,13 +1,12 @@
-"""These classes parse information from Clients into a universal,
-downloadable form.
-"""
+"""These classes parse information from Clients into a universal, downloadable form."""
+
+from __future__ import annotations
 
-import functools
 import logging
 import os
 import re
 from tempfile import gettempdir
-from typing import Generator, Iterable, Union
+from typing import Dict, Generator, Iterable, Optional, Union
 
 import click
 from pathvalidate import sanitize_filename
@@ -26,12 +25,7 @@ from .utils import (
     tqdm_download,
 )
 
-logger = logging.getLogger(__name__)
-
-TYPE_REGEXES = {
-    "remaster": re.compile(r"(?i)(re)?master(ed)?"),
-    "extra": re.compile(r"(?i)(anniversary|deluxe|live|collector|demo|expanded|remix)"),
-}
+logger = logging.getLogger("streamrip")
 
 
 class Album(Tracklist):
@@ -57,7 +51,11 @@ class Album(Tracklist):
 
         self.sampling_rate = None
         self.bit_depth = None
-        self.container = None
+        self.container: Optional[str] = None
+
+        self.disctotal: int
+        self.tracktotal: int
+        self.albumartist: str
 
         # usually an unpacked TrackMetadata.asdict()
         self.__dict__.update(kwargs)
@@ -71,7 +69,6 @@ class Album(Tracklist):
 
     def load_meta(self):
         """Load detailed metadata from API using the id."""
-
         assert hasattr(self, "id"), "id must be set to load metadata"
         resp = self.client.get(self.id, media_type="album")
 
@@ -87,6 +84,13 @@ class Album(Tracklist):
 
     @classmethod
     def from_api(cls, resp: dict, client: Client):
+        """Create an Album object from an API response.
+
+        :param resp:
+        :type resp: dict
+        :param client:
+        :type client: Client
+        """
         if client.source == "soundcloud":
             return Playlist.from_api(resp, client)
 
@@ -94,9 +98,17 @@ class Album(Tracklist):
         return cls(client, **info.asdict())
 
     def _prepare_download(self, **kwargs):
+        """Prepare the download of the album.
+
+        :param kwargs:
+        """
         # Generate the folder name
         self.folder_format = kwargs.get("folder_format", FOLDER_FORMAT)
-        self.quality = min(kwargs.get("quality", 3), self.client.max_quality)
+        if not hasattr(self, "quality"):
+            self.quality = min(
+                kwargs.get("quality", 3), self.client.max_quality
+            )
+
         self.folder = self._get_formatted_folder(
             kwargs.get("parent_folder", "StreamripDownloads"), self.quality
         )
@@ -117,15 +129,14 @@ class Album(Tracklist):
         if embed_cover_url is not None:
             tqdm_download(embed_cover_url, cover_path)
         else:  # sometimes happens with Deezer
-            cover_url = functools.reduce(
-                lambda c1, c2: c1 or c2, self.cover_urls.values()
-            )
+            cover_url = [u for u in self.cover_urls.values() if u][0]
             tqdm_download(cover_url, cover_path)
 
-        if kwargs.get("keep_hires_cover", True):
-            tqdm_download(
-                self.cover_urls["original"], os.path.join(self.folder, "cover.jpg")
-            )
+        hires_cov_path = os.path.join(self.folder, "cover.jpg")
+        if kwargs.get("keep_hires_cover", True) and not os.path.exists(
+            hires_cov_path
+        ):
+            tqdm_download(self.cover_urls["original"], hires_cov_path)
 
         cover_size = os.path.getsize(cover_path)
         if cover_size > FLAC_MAX_BLOCKSIZE:  # 16.77 MB
@@ -146,37 +157,59 @@ class Album(Tracklist):
             self.cover_obj = None
 
         # Download the booklet if applicable
-        if self.get("booklets") and kwargs.get("download_booklets", True):
+        if (
+            self.get("booklets")
+            and kwargs.get("download_booklets", True)
+            and not any(f.endswith(".pdf") for f in os.listdir(self.folder))
+        ):
             click.secho("\nDownloading booklets", fg="blue")
             for item in self.booklets:
                 Booklet(item).download(parent_folder=self.folder)
 
-    def _download_item(
+    def _download_item(  # type: ignore
         self,
         track: Union[Track, Video],
         quality: int = 3,
         database: MusicDB = None,
         **kwargs,
     ) -> bool:
+        """Download an item.
+
+        :param track: The item.
+        :type track: Union[Track, Video]
+        :param quality:
+        :type quality: int
+        :param database:
+        :type database: MusicDB
+        :param kwargs:
+        :rtype: bool
+        """
         logger.debug("Downloading track to %s", self.folder)
         if self.disctotal > 1 and isinstance(track, Track):
-            disc_folder = os.path.join(self.folder, f"Disc {track.meta.discnumber}")
+            disc_folder = os.path.join(
+                self.folder, f"Disc {track.meta.discnumber}"
+            )
             kwargs["parent_folder"] = disc_folder
         else:
             kwargs["parent_folder"] = self.folder
 
-        if not track.download(quality=quality, database=database, **kwargs):
+        if not track.download(
+            quality=min(self.quality, quality), database=database, **kwargs
+        ):
             return False
 
         logger.debug("tagging tracks")
         # deezer tracks come tagged
         if kwargs.get("tag_tracks", True) and self.client.source != "deezer":
-            track.tag(cover=self.cover_obj, embed_cover=kwargs.get("embed_cover", True))
+            track.tag(
+                cover=self.cover_obj,
+                embed_cover=kwargs.get("embed_cover", True),
+            )
 
         return True
 
     @staticmethod
-    def _parse_get_resp(resp: dict, client: Client) -> dict:
+    def _parse_get_resp(resp: dict, client: Client) -> TrackMetadata:
         """Parse information from a client.get(query, 'album') call.
 
         :param resp:
@@ -188,8 +221,7 @@ class Album(Tracklist):
         return meta
 
     def _load_tracks(self, resp):
-        """Given an album metadata dict returned by the API, append all of its
-        tracks to `self`.
+        """Load the tracks into self from an API response.
 
         This uses a classmethod to convert an item into a Track object, which
         stores the metadata inside a TrackMetadata object.
@@ -197,22 +229,19 @@ class Album(Tracklist):
         logging.debug(f"Loading {self.tracktotal} tracks to album")
         for track in _get_tracklist(resp, self.client.source):
             if track.get("type") == "Music Video":
-                self.append(
-                    Video.from_album_meta(
-                        track,
-                        self.client,
-                    )
-                )
+                self.append(Video.from_album_meta(track, self.client))
             else:
                 self.append(
                     Track.from_album_meta(
-                        album=self.meta,
-                        track=track,
-                        client=self.client,
+                        album=self.meta, track=track, client=self.client
                     )
                 )
 
     def _get_formatter(self) -> dict:
+        """Get a formatter that is used for previews in core.py.
+
+        :rtype: dict
+        """
         fmt = dict()
         for key in ALBUM_KEYS:
             # default to None
@@ -227,22 +256,41 @@ class Album(Tracklist):
         return fmt
 
     def _get_formatted_folder(self, parent_folder: str, quality: int) -> str:
+        """Generate the folder name for this album.
+
+        :param parent_folder:
+        :type parent_folder: str
+        :param quality:
+        :type quality: int
+        :rtype: str
+        """
         # necessary to format the folder
         self.container = get_container(quality, self.client.source)
         if self.container in ("AAC", "MP3"):
             # lossy codecs don't have these metrics
             self.bit_depth = self.sampling_rate = None
 
-        formatted_folder = clean_format(self.folder_format, self._get_formatter())
+        formatted_folder = clean_format(
+            self.folder_format, self._get_formatter()
+        )
 
         return os.path.join(parent_folder, formatted_folder)
 
     @property
     def title(self) -> str:
+        """Get the title of the album.
+
+        :rtype: str
+        """
         return self.album
 
     @title.setter
     def title(self, val: str):
+        """Set the title of the Album.
+
+        :param val:
+        :type val: str
+        """
         self.album = val
 
     def __repr__(self) -> str:
@@ -257,17 +305,21 @@ class Album(Tracklist):
         return f"<Album: V/A - {self.title}>"
 
     def __str__(self) -> str:
-        """Return a readable string representation of
-        this album.
+        """Return a readable string representation of this album.
 
         :rtype: str
         """
         return f"{self['albumartist']} - {self['title']}"
 
     def __len__(self) -> int:
+        """Get the length of the album.
+
+        :rtype: int
+        """
         return self.tracktotal
 
     def __hash__(self):
+        """Hash the album."""
         return hash(self.id)
 
 
@@ -302,8 +354,7 @@ class Playlist(Tracklist):
 
     @classmethod
     def from_api(cls, resp: dict, client: Client):
-        """Return a Playlist object initialized with information from
-        a search result returned by the API.
+        """Return a Playlist object from an API response.
 
         :param resp: a single search result entry of a playlist
         :type resp: dict
@@ -326,7 +377,7 @@ class Playlist(Tracklist):
         self.loaded = True
 
     def _load_tracks(self, new_tracknumbers: bool = True):
-        """Parses the tracklist returned by the API.
+        """Parse the tracklist returned by the API.
 
         :param new_tracknumbers: replace tracknumber tag with playlist position
         :type new_tracknumbers: bool
@@ -334,7 +385,9 @@ class Playlist(Tracklist):
         if self.client.source == "qobuz":
             self.name = self.meta["name"]
             self.image = self.meta["images"]
-            self.creator = safe_get(self.meta, "owner", "name", default="Qobuz")
+            self.creator = safe_get(
+                self.meta, "owner", "name", default="Qobuz"
+            )
 
             tracklist = self.meta["tracks"]["items"]
 
@@ -347,7 +400,9 @@ class Playlist(Tracklist):
         elif self.client.source == "tidal":
             self.name = self.meta["title"]
             self.image = tidal_cover_url(self.meta["image"], 640)
-            self.creator = safe_get(self.meta, "creator", "name", default="TIDAL")
+            self.creator = safe_get(
+                self.meta, "creator", "name", default="TIDAL"
+            )
 
             tracklist = self.meta["tracks"]
 
@@ -364,7 +419,9 @@ class Playlist(Tracklist):
         elif self.client.source == "deezer":
             self.name = self.meta["title"]
             self.image = self.meta["picture_big"]
-            self.creator = safe_get(self.meta, "creator", "name", default="Deezer")
+            self.creator = safe_get(
+                self.meta, "creator", "name", default="Deezer"
+            )
 
             tracklist = self.meta["tracks"]
 
@@ -407,18 +464,20 @@ class Playlist(Tracklist):
 
         logger.debug(f"Loaded {len(self)} tracks from playlist {self.name}")
 
-    def _prepare_download(self, parent_folder: str = "StreamripDownloads", **kwargs):
+    def _prepare_download(
+        self, parent_folder: str = "StreamripDownloads", **kwargs
+    ):
         fname = sanitize_filename(self.name)
         self.folder = os.path.join(parent_folder, fname)
 
         self.__download_index = 1  # used for tracknumbers
         self.download_message()
 
-    def _download_item(self, item: Track, **kwargs):
-        kwargs['parent_folder'] = self.folder
+    def _download_item(self, item: Track, **kwargs) -> bool:  # type: ignore
+        kwargs["parent_folder"] = self.folder
         if self.client.source == "soundcloud":
             item.load_meta()
-            click.secho(f"Downloading {item!s}", fg='blue')
+            click.secho(f"Downloading {item!s}", fg="blue")
 
         if kwargs.get("set_playlist_to_album", False):
             item["album"] = self.name
@@ -438,8 +497,7 @@ class Playlist(Tracklist):
 
     @staticmethod
     def _parse_get_resp(item: dict, client: Client) -> dict:
-        """Parses information from a search result returned
-        by a client.search call.
+        """Parse information from a search result returned by a client.search call.
 
         :param item:
         :type item: dict
@@ -474,6 +532,10 @@ class Playlist(Tracklist):
 
     @property
     def title(self) -> str:
+        """Get the title.
+
+        :rtype: str
+        """
         return self.name
 
     def __repr__(self) -> str:
@@ -484,8 +546,7 @@ class Playlist(Tracklist):
         return f"<Playlist: {self.name}>"
 
     def __str__(self) -> str:
-        """Return a readable string representation of
-        this track.
+        """Return a readable string representation of this track.
 
         :rtype: str
         """
@@ -529,13 +590,18 @@ class Artist(Tracklist):
 
     # override
     def download(self, **kwargs):
+        """Download all items in self.
+
+        :param kwargs:
+        """
         iterator = self._prepare_download(**kwargs)
         for item in iterator:
             self._download_item(item, **kwargs)
 
     def _load_albums(self):
-        """From the discography returned by client.get(query, 'artist'),
-        generate album objects and append them to self.
+        """Load Album objects to self.
+
+        This parses the response of client.get(query, 'artist') responses.
         """
         if self.client.source == "qobuz":
             self.name = self.meta["name"]
@@ -557,15 +623,28 @@ class Artist(Tracklist):
             self.append(Album.from_api(album, self.client))
 
     def _prepare_download(
-        self, parent_folder: str = "StreamripDownloads", filters: tuple = (), **kwargs
+        self,
+        parent_folder: str = "StreamripDownloads",
+        filters: tuple = (),
+        **kwargs,
     ) -> Iterable:
+        """Prepare the download.
+
+        :param parent_folder:
+        :type parent_folder: str
+        :param filters:
+        :type filters: tuple
+        :param kwargs:
+        :rtype: Iterable
+        """
         folder = sanitize_filename(self.name)
-        folder = os.path.join(parent_folder, folder)
+        self.folder = os.path.join(parent_folder, folder)
 
         logger.debug("Artist folder: %s", folder)
         logger.debug(f"Length of tracklist {len(self)}")
         logger.debug(f"Filters: {filters}")
 
+        final: Iterable
         if "repeats" in filters:
             final = self._remove_repeats(bit_depth=max, sampling_rate=min)
             filters = tuple(f for f in filters if f != "repeats")
@@ -573,14 +652,16 @@ class Artist(Tracklist):
             final = self
 
         if isinstance(filters, tuple) and self.client.source == "qobuz":
-            filter_funcs = (getattr(self, f"_{filter_}") for filter_ in filters)
+            filter_funcs = (
+                getattr(self, f"_{filter_}") for filter_ in filters
+            )
             for func in filter_funcs:
                 final = filter(func, final)
 
         self.download_message()
         return final
 
-    def _download_item(
+    def _download_item(  # type: ignore
         self,
         item,
         parent_folder: str = "StreamripDownloads",
@@ -588,15 +669,27 @@ class Artist(Tracklist):
         database: MusicDB = None,
         **kwargs,
     ) -> bool:
+        """Download an item.
+
+        :param item:
+        :param parent_folder:
+        :type parent_folder: str
+        :param quality:
+        :type quality: int
+        :param database:
+        :type database: MusicDB
+        :param kwargs:
+        :rtype: bool
+        """
         try:
             item.load_meta()
         except NonStreamable:
             logger.info("Skipping album, not available to stream.")
-            return
+            return False
 
         # always an Album
         status = item.download(
-            parent_folder=parent_folder,
+            parent_folder=self.folder,
             quality=quality,
             database=database,
             **kwargs,
@@ -605,12 +698,17 @@ class Artist(Tracklist):
 
     @property
     def title(self) -> str:
+        """Get the artist name.
+
+        Implemented for consistency.
+
+        :rtype: str
+        """
         return self.name
 
     @classmethod
     def from_api(cls, item: dict, client: Client, source: str = "qobuz"):
-        """Create an Artist object from the api response of Qobuz, Tidal,
-        or Deezer.
+        """Create an Artist object from the api response of Qobuz, Tidal, or Deezer.
 
         :param resp: response dict
         :type resp: dict
@@ -649,14 +747,22 @@ class Artist(Tracklist):
 
     # ----------- Filters --------------
 
+    TYPE_REGEXES = {
+        "remaster": re.compile(r"(?i)(re)?master(ed)?"),
+        "extra": re.compile(
+            r"(?i)(anniversary|deluxe|live|collector|demo|expanded|remix)"
+        ),
+    }
+
     def _remove_repeats(self, bit_depth=max, sampling_rate=max) -> Generator:
-        """Remove the repeated albums from self. May remove different
-        versions of the same album.
+        """Remove the repeated albums from self.
+
+        May remove different versions of the same album.
 
         :param bit_depth: either max or min functions
         :param sampling_rate: either max or min functions
         """
-        groups = dict()
+        groups: Dict[str, list] = {}
         for album in self:
             if (t := self.essence(album.title)) not in groups:
                 groups[t] = []
@@ -667,14 +773,15 @@ class Artist(Tracklist):
             best_bd = bit_depth(a["bit_depth"] for a in group)
             best_sr = sampling_rate(a["sampling_rate"] for a in group)
             for album in group:
-                if album["bit_depth"] == best_bd and album["sampling_rate"] == best_sr:
+                if (
+                    album["bit_depth"] == best_bd
+                    and album["sampling_rate"] == best_sr
+                ):
                     yield album
                     break
 
     def _non_studio_albums(self, album: Album) -> bool:
-        """Passed as a parameter by the user.
-
-        This will download only studio albums.
+        """Filter non-studio-albums.
 
         :param artist: usually self
         :param album: the album to check
@@ -683,11 +790,11 @@ class Artist(Tracklist):
         """
         return (
             album["albumartist"] != "Various Artists"
-            and TYPE_REGEXES["extra"].search(album.title) is None
+            and self.TYPE_REGEXES["extra"].search(album.title) is None
         )
 
     def _features(self, album: Album) -> bool:
-        """Passed as a parameter by the user.
+        """Filter features.
 
         This will download only albums where the requested
         artist is the album artist.
@@ -700,39 +807,34 @@ class Artist(Tracklist):
         return self["name"] == album["albumartist"]
 
     def _extras(self, album: Album) -> bool:
-        """Passed as a parameter by the user.
-
-        This will skip any extras.
+        """Filter extras.
 
         :param artist: usually self
         :param album: the album to check
         :type album: Album
         :rtype: bool
         """
-        return TYPE_REGEXES["extra"].search(album.title) is None
+        return self.TYPE_REGEXES["extra"].search(album.title) is None
 
     def _non_remasters(self, album: Album) -> bool:
-        """Passed as a parameter by the user.
-
-        This will download only remasterd albums.
+        """Filter non remasters.
 
         :param artist: usually self
         :param album: the album to check
         :type album: Album
         :rtype: bool
         """
-        return TYPE_REGEXES["remaster"].search(album.title) is not None
+        return self.TYPE_REGEXES["remaster"].search(album.title) is not None
 
     def _non_albums(self, album: Album) -> bool:
-        """This will ignore non-album releases.
+        """Filter releases that are not albums.
 
         :param artist: usually self
         :param album: the album to check
         :type album: Album
         :rtype: bool
         """
-        # Doesn't work yet
-        return album["release_type"] == "album"
+        return len(album) > 1
 
     # --------- Magic Methods --------
 
@@ -744,19 +846,22 @@ class Artist(Tracklist):
         return f"<Artist: {self.name}>"
 
     def __str__(self) -> str:
-        """Return a readable string representation of
-        this Artist.
+        """Return a readable string representation of this Artist.
 
         :rtype: str
         """
         return self.name
 
-    def __hash__(self) -> int:
+    def __hash__(self):
+        """Hash self."""
         return hash(self.id)
 
 
 class Label(Artist):
+    """Represents a downloadable Label."""
+
     def load_meta(self):
+        """Load metadata given an id."""
         assert self.client.source == "qobuz", "Label source must be qobuz"
 
         resp = self.client.get(self.id, "label")
@@ -767,11 +872,11 @@ class Label(Artist):
         self.loaded = True
 
     def __repr__(self):
+        """Return a string representation of the Label."""
         return f"<Label - {self.name}>"
 
     def __str__(self) -> str:
-        """Return a readable string representation of
-        this track.
+        """Return the name of the Label.
 
         :rtype: str
         """
@@ -782,7 +887,7 @@ class Label(Artist):
 
 
 def _get_tracklist(resp: dict, source: str) -> list:
-    """Returns the tracklist from an API response.
+    """Return the tracklist from an API response.
 
     :param resp:
     :type resp: dict
